/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.osi;

import java.util.Objects;

import org.apache.commons.lang3.StringUtils;
import org.apache.hc.client5.http.auth.AuthScope;
import org.apache.hc.client5.http.auth.CredentialsProvider;
import org.apache.hc.client5.http.auth.UsernamePasswordCredentials;
import org.apache.hc.client5.http.impl.auth.BasicCredentialsProvider;
import org.apache.hc.client5.http.impl.classic.HttpClientBuilder;
import org.apache.hc.client5.http.impl.classic.HttpClients;
import org.apache.hc.core5.http.HttpHost;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.boot.web.client.RestTemplateCustomizer;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Profile;
import org.springframework.http.client.HttpComponentsClientHttpRequestFactory;
import org.springframework.web.client.RestTemplate;

@Profile("!e2e")
@Configuration
public class PostfachConfiguration implements RestTemplateCustomizer {

	static final String HEADER_API_KEY = "apiKey";
	static final String HEADER_API_REALM = "realm";

	@Autowired
	private OsiPostfachProperties properties;

	@Bean
	@ConditionalOnProperty(prefix = OsiPostfachProperties.PROXY_API_PREFIX, name = { "url", "key", "realm" })
	public RestTemplate restTemplate(RestTemplateBuilder builder, OsiPostfachProperties properties) {
		return builder
				.defaultHeader(HEADER_API_KEY, properties.getKey())
				.defaultHeader(HEADER_API_REALM, properties.getRealm())
				.build();
	}

	@Override
	public void customize(RestTemplate restTemplate) {
		if (StringUtils.isNotBlank(properties.getProxyConfiguration().getHost())) {
			setProxy(restTemplate);
		}
	}

	private void setProxy(RestTemplate restTemplate) {
		var proxyConfig = properties.getProxyConfiguration();
		var port = Objects.isNull(proxyConfig.getPort()) ? -1 : proxyConfig.getPort();
		var proxy = new HttpHost(proxyConfig.getHost(), port);

		var httpClient = HttpClientBuilder.create()
				.setProxy(proxy).setDefaultCredentialsProvider(buildCredentialsProvider(proxy))
				.build();

		HttpClients.custom().build();
		restTemplate.setRequestFactory(new HttpComponentsClientHttpRequestFactory(httpClient));
	}

	private CredentialsProvider buildCredentialsProvider(HttpHost proxyHost) {
		var proxyConfig = properties.getProxyConfiguration();

		if (!proxyConfig.isAuthenticationRequired()) {
			return null;
		}

		var credsProvider = new BasicCredentialsProvider();
		credsProvider.setCredentials(
				new AuthScope(proxyHost),
				new UsernamePasswordCredentials(proxyConfig.getUsername(), proxyConfig.getPassword().toCharArray()));
		return credsProvider;
	}

}
