package de.ozgcloud.nachrichten.postfach.bayernid;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.time.ZoneId;
import java.time.ZonedDateTime;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.Spy;

import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcAttachmentMetadata;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcAttachments;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcBayernIdMessageMetadata;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcSendBayernIdMessageRequest;
import de.ozgcloud.nachrichten.postfach.osi.MessageTestFactory;

class BayernIdPostfachNachrichtMapperTest {

	@Spy
	private BayernIdPostfachNachrichtMapper mapper = Mappers.getMapper(BayernIdPostfachNachrichtMapper.class);

	@Nested
	class TestMapToSendBayernIdMessageMetadataRequest {

		private static final PostfachNachricht NACHRICHT = PostfachNachrichtTestFactory.create();
		private static final Absender ABSENDER = AbsenderTestFactory.create();

		@Test
		void shouldIgnoreAttachments() {
			var request = mapper.toSendBayernIdMessageMetadataRequest(NACHRICHT, ABSENDER);

			assertThat(request.getAttachments()).isEqualTo(GrpcAttachments.getDefaultInstance());
		}

		@Test
		void shouldCallToBayernIdMessageMetadata() {
			mapper.toSendBayernIdMessageMetadataRequest(NACHRICHT, ABSENDER);

			verify(mapper).toBayernIdMessageMetadata(NACHRICHT, ABSENDER);
		}

		@Nested
		class TestMapToBayernIdMessageMetadata {

			@Test
			void shouldSetMessageId() {
				var metadata = toMessageMetadata();

				assertThat(metadata.getMessageId()).isEqualTo(MessageTestFactory.MESSAGE_ID);
			}

			@Nested
			class TestCreatedAt {

				@Test
				void shouldCallConvertZonedDateTime() {
					toMessageMetadata();

					verify(mapper).convertZonedDateTime(PostfachNachrichtTestFactory.CREATED_AT);
				}

				@Test
				void shouldSetCreatedAt() {
					doReturn(PostfachNachrichtTestFactory.CREATED_AT_STR).when(mapper).convertZonedDateTime(any());

					var metadata = toMessageMetadata();

					assertThat(metadata.getCreatedAt()).isEqualTo(PostfachNachrichtTestFactory.CREATED_AT_STR);
				}
			}

			@Test
			void shouldSetSubject() {
				var metadata = toMessageMetadata();

				assertThat(metadata.getSubject()).isEqualTo(MessageTestFactory.SUBJECT);
			}

			@Test
			void shouldSetStorkQaaLevel() {
				var metadata = toMessageMetadata();

				assertThat(metadata.getStorkQaaLevel()).isEqualTo(BayernIdPostfachNachrichtMapper.MESSAGE_LEVEL);
			}

			@Test
			void shouldSetVorgangId() {
				var metadata = toMessageMetadata();

				assertThat(metadata.getVorgangId()).isEqualTo(MessageTestFactory.VORGANG_ID);
			}

			@Test
			void shouldSetText() {
				var metadata = toMessageMetadata();

				assertThat(metadata.getText()).isEqualTo(PostfachNachrichtTestFactory.MAIL_BODY);
			}

			@Nested
			class TestMapAbsender {

				@Test
				void shouldSetName() {
					var metadata = toMessageMetadata();

					assertThat(metadata.getAbsender().getName()).isEqualTo(AbsenderTestFactory.NAME);
				}

				@Test
				void shouldSetAnschrift() {
					var metadata = toMessageMetadata();

					assertThat(metadata.getAbsender().getAnschrift()).isEqualTo(AbsenderTestFactory.ANSCHRIFT);
				}

				@Test
				void shouldSetDienst() {
					var metadata = toMessageMetadata();

					assertThat(metadata.getAbsender().getDienst()).isEqualTo(AbsenderTestFactory.DIENST);
				}

				@Test
				void shouldSetMandant() {
					var metadata = toMessageMetadata();

					assertThat(metadata.getAbsender().getMandant()).isEqualTo(AbsenderTestFactory.MANDANT);
				}

				@Test
				void shouldSetGemeindeschluessel() {
					var metadata = toMessageMetadata();

					assertThat(metadata.getAbsender().getGemeindeschluessel()).isEqualTo(AbsenderTestFactory.GEMEINDE_SCHLUESSEL);
				}

			}

			@Nested
			class TestMapToEmpfaenger {

				@Test
				void shouldSetPostkorbId() {
					var metadata = toMessageMetadata();

					assertThat(metadata.getEmpfaenger().getPostkorbId()).isEqualTo(MessageTestFactory.POSTFACH_ID);
				}
			}

			private GrpcBayernIdMessageMetadata toMessageMetadata() {
				return mapper.toBayernIdMessageMetadata(NACHRICHT, ABSENDER);
			}
		}
	}

	@Nested
	class TestMapToSendBayernIdAttachmentsMetadata {

		private static final BayernIdAttachment ATTACHMENT = BayernIdAttachmentTestFactory.create();

		@Test
		void shouldIgnoreMessageMetadata() {
			var request = toMessageRequest();

			assertThat(request.getMessageMetadata()).isEqualTo(GrpcBayernIdMessageMetadata.getDefaultInstance());
		}

		@Test
		void shouldIgnoreContent() {
			var request = toMessageRequest();

			assertThat(request.getAttachments().getContent()).isEmpty();
		}

		@Test
		void shouldCallToAttachmentsMetadata() {
			toMessageRequest();

			verify(mapper).toAttachmentMetadata(ATTACHMENT);
		}

		private GrpcSendBayernIdMessageRequest toMessageRequest() {
			return mapper.toSendBayernIdAttachmentsMetadataRequest(ATTACHMENT);
		}
	}

	@Nested
	class TestMapToAttachmentMetadata {

		@Test
		void shouldSetFileName() {
			var metadata = toMessageRequest();

			assertThat(metadata.getFileName()).isEqualTo(BayernIdAttachmentTestFactory.FILENAME);
		}

		@Test
		void shouldSetFileType() {
			var metadata = toMessageRequest();

			assertThat(metadata.getFileType()).isEqualTo(BayernIdAttachmentTestFactory.CONTENT_TYPE);
		}

		private GrpcAttachmentMetadata toMessageRequest() {
			return mapper.toAttachmentMetadata(BayernIdAttachmentTestFactory.create());
		}
	}

	@Nested
	class TestMapToGregorianCalendar {

		@Test
		void shouldReturnGregorianCalendar() {
			var expectedDateTiem = "2024-02-14T23:29:25.000+01:00";
			var dateTiem = ZonedDateTime.of(2024, 2, 14, 23, 29, 25, 0, ZoneId.of("Europe/Berlin"));

			var calendar = mapper.convertZonedDateTime(dateTiem);

			assertThat(calendar).isEqualTo(expectedDateTiem);
		}
	}

	@Nested
	class TestMapFromSendBayernIdMessageResponse {

		@Test
		void shouldSetSuccess() {
			var result = mapper.fromSendBayernIdMessageResponse(GrpcSendBayernIdMessageResponseTestFactory.create());

			assertThat(result.isSuccess()).isTrue();
		}

		@Test
		void shouldSetStatus() {
			var result = mapper.fromSendBayernIdMessageResponse(GrpcSendBayernIdMessageResponseTestFactory.create());

			assertThat(result.getStatus()).isEqualTo(GrpcSendBayernIdMessageResponseTestFactory.STATUS);
		}

		@Test
		void shouldSetMessage(){
			var result = mapper.fromSendBayernIdMessageResponse(GrpcSendBayernIdMessageResponseTestFactory.create());

			assertThat(result.getMessage()).isEqualTo(GrpcSendBayernIdMessageResponseTestFactory.MESSAGE_TEXT);
		}
	}

	@Nested
	class TestSkipNulls {

		@Test
		void shouldNotSetNull() {
			var bayernIdMessageMetadata = mapper.toBayernIdMessageMetadata(PostfachNachrichtTestFactory.createBuilder().messageId(null).build(),
					AbsenderTestFactory.create());

			assertThat(bayernIdMessageMetadata.getMessageId()).isEmpty();
		}

		@Test
		void shouldReturnTrueIfNotNull() {
			var result = mapper.nonNull("test");

			assertThat(result).isTrue();
		}
	}
}
