/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.bayernid;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeoutException;
import java.util.function.BiFunction;
import java.util.function.Function;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.postfach.FileId;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;
import de.ozgcloud.nachrichten.postfach.PostfachRuntimeException;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.BayernIdProxyServiceGrpc;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcSendBayernIdMessageRequest;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcSendBayernIdMessageResponse;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;
import lombok.SneakyThrows;

class BayernIdPostfachRemoteServiceTest {

	@Spy
	@InjectMocks
	private BayernIdPostfachRemoteService service;

	@Mock
	private BayernIdProxyServiceGrpc.BayernIdProxyServiceStub bayernIdProxyServiceStub;
	@Mock
	private BayernIdProperties properties;
	@Mock
	private BayernIdPostfachResponseHandler responseHandler;
	@Mock
	private BayernIdPostfachNachrichtMapper mapper;
	@Mock
	private BayernIdAttachmentService attachmentService;

	@Nested
	@DisplayName("Get all messages")
	class TestGetAllMessages {

		@Test
		void shouldThrowUnsupportedOperationException() {
			assertThatExceptionOfType(UnsupportedOperationException.class).isThrownBy(service::getAllMessages);
		}
	}

	@Nested
	@DisplayName("Delete messages")
	class TestDeleteMessages {

		@Test
		void shouldThrowUnsupportedOperationException() {
			assertThatExceptionOfType(UnsupportedOperationException.class).isThrownBy(() -> service.deleteMessage(StringUtils.EMPTY));
		}
	}

	@Nested
	@DisplayName("Send message")
	class TestSendMessage {

		@Mock
		private GrpcSendBayernIdMessageResponse grpcResponse;
		@Mock
		private MessageWithFilesSender messageWithFilesSender;
		@Mock
		private BayernIdResponse bayernIdResponse;

		@Nested
		class TestSend {

			@BeforeEach
			void setup() {
				doReturn(grpcResponse).when(service).waitUntilTransferCompleted(any());
			}

			@Test
			void shouldCallCreateMessageWithFileSender() {
				sendMessage();

				verify(service).createMessageWithFilesSender(any());
			}

			@Test
			void shouldCallSend() {
				doReturn(messageWithFilesSender).when(service).createMessageWithFilesSender(any());

				sendMessage();

				verify(messageWithFilesSender).send();
			}

			@Test
			void shouldCallWaitForResults() {
				var expectedSender = mock(MessageWithFilesSender.class);
				when(messageWithFilesSender.send()).thenReturn(expectedSender);
				doReturn(messageWithFilesSender).when(service).createMessageWithFilesSender(any());

				sendMessage();

				verify(service).waitUntilTransferCompleted(expectedSender);
			}

			@Test
			void shouldCallMaapper() {
				sendMessage();

				verify(mapper).fromSendBayernIdMessageResponse(grpcResponse);
			}

			@Test
			void shouldCallResponseHandler() {
				when(mapper.fromSendBayernIdMessageResponse(any())).thenReturn(bayernIdResponse);

				sendMessage();

				verify(responseHandler).handleResponse(bayernIdResponse);
			}
		}

		@Nested
		class TestExceptionHandling {

			@Test
			void shouldHandleExceptionWhenSending() {
				var thrownException = new RuntimeException("Test");
				when(messageWithFilesSender.send()).thenThrow(thrownException);
				doReturn(messageWithFilesSender).when(service).createMessageWithFilesSender(any());

				assertThrows(PostfachRuntimeException.class, TestSendMessage.this::sendMessage);
			}

			@Test
			void shouldHandleExceptionWhenWaiting() {
				var thrownException = new RuntimeException("Test");
				doThrow(thrownException).when(service).waitUntilTransferCompleted(any());

				assertThrows(PostfachRuntimeException.class, TestSendMessage.this::sendMessage);
			}

			@Test
			void shouldPassBayernIdServerException() {
				doReturn(grpcResponse).when(service).waitUntilTransferCompleted(any());
				var thrownException = new BayernIdServerException("Test");
				doThrow(thrownException).when(responseHandler).handleResponse(any());

				var resultException = assertThrows(BayernIdServerException.class, TestSendMessage.this::sendMessage);
				assertThat(resultException).isEqualTo(thrownException);
			}
		}

		private void sendMessage() {
			service.sendMessage(PostfachNachrichtTestFactory.create());
		}
	}

	@Nested
	class TestCreateMessageWithFileSender {

		@Mock
		private Function<StreamObserver<GrpcSendBayernIdMessageResponse>, CallStreamObserver<GrpcSendBayernIdMessageRequest>> streamObserverBuilder;
		@Mock
		private GrpcSendBayernIdMessageRequest metadataRequest;
		@Mock
		private Function<String, BayernIdAttachment> attachmentBuilder;
		@Mock
		private Function<BayernIdAttachment, GrpcSendBayernIdMessageRequest> attachmentMetadataMapper;
		@Mock
		private BiFunction<byte[], Integer, GrpcSendBayernIdMessageRequest> chunkBuilder;

		@Test
		void shouldSetStreamObserverBuilder() {
			doReturn(streamObserverBuilder).when(service).buildCallStreamObserverBuilder();

			var sender = createMessageWithFilesSender();

			assertThat(sender).extracting("reqObserverBuilder").isEqualTo(streamObserverBuilder);
		}

		@Test
		void shouldCallMessageMetadataMapper() {
			var absender = AbsenderTestFactory.create();
			when(properties.getAbsender()).thenReturn(absender);
			var nachricht = PostfachNachrichtTestFactory.create();

			service.createMessageWithFilesSender(nachricht);

			verify(mapper).toSendBayernIdMessageMetadataRequest(nachricht, absender);
		}

		@Test
		void shouldSetMessageMetadata() {
			when(mapper.toSendBayernIdMessageMetadataRequest(any(), any())).thenReturn(metadataRequest);

			var sender = createMessageWithFilesSender();

			assertThat(sender).extracting("messageMetadata").isEqualTo(metadataRequest);
		}

		@Test
		void shouldSetAttachmentIds() {
			var nachricht = PostfachNachrichtTestFactory.create();

			var sender = createMessageWithFilesSender();

			assertThat(sender).extracting("attachmentIds").isEqualTo(nachricht.getAttachments());
		}

		@Test
		void shouldSetAttachmentBuilder() {
			doReturn(attachmentBuilder).when(service).buildAttachmentBuilder();

			var sender = createMessageWithFilesSender();

			assertThat(sender).extracting("toAttachment").isEqualTo(attachmentBuilder);
		}

		@Test
		void shouldSetAttachmentMetadataBuilder() {
			doReturn(attachmentMetadataMapper).when(service).buildAttachmentMetadataMapper();

			var sender = createMessageWithFilesSender();

			assertThat(sender).extracting("attachmentMetadataMapper").isEqualTo(attachmentMetadataMapper);
		}

		@Test
		void shouldSetChunkBuilder() {
			doReturn(chunkBuilder).when(service).buildChunkRequest();

			var sender = createMessageWithFilesSender();

			assertThat(sender).extracting("chunkBuilder").isEqualTo(chunkBuilder);
		}

		private MessageWithFilesSender createMessageWithFilesSender() {
			return service.createMessageWithFilesSender(PostfachNachrichtTestFactory.create());
		}
	}

	@Nested
	class TestBuildCallStreamObserverBuilder {

		@Mock
		private CallStreamObserver<GrpcSendBayernIdMessageRequest> requestObserver;
		@Mock
		private StreamObserver<GrpcSendBayernIdMessageResponse> responseObserver;

		@Test
		void shouldCallSendMessageAsStream() {
			service.buildCallStreamObserverBuilder().apply(responseObserver);

			verify(bayernIdProxyServiceStub).sendMessageAsStream(responseObserver);
		}

		@Test
		void shouldReturnRequestObserver() {
			when(bayernIdProxyServiceStub.sendMessageAsStream(any())).thenReturn(requestObserver);

			var result = service.buildCallStreamObserverBuilder().apply(responseObserver);

			assertThat(result).isEqualTo(requestObserver);
		}
	}

	@Nested
	class TestBuildAttachmentBuilder {

		@Captor
		private ArgumentCaptor<FileId> attachmentIdCaptor;

		@Test
		void shouldCallGetMessageAttachment (){
			var attachmentId = "test";

			service.buildAttachmentBuilder().apply(attachmentId);

			verify(attachmentService).getMessageAttachment(attachmentIdCaptor.capture());
			assertThat(attachmentIdCaptor.getValue()).extracting(FileId::toString).isEqualTo(attachmentId);
		}

		@Test
		void shouldReturnAttachment() {
			var attachment = BayernIdAttachmentTestFactory.create();
			when(attachmentService.getMessageAttachment(any())).thenReturn(attachment);

			var result = service.buildAttachmentBuilder().apply(StringUtils.EMPTY);

			assertThat(result).isEqualTo(attachment);
		}
	}

	@Nested
	class TestBuildChunkRequest {

		@Captor
		private ArgumentCaptor<byte[]> bytesCaptor;
		@Captor
		private ArgumentCaptor<Integer> lengthCaptor;

		@Test
		void shouldReturnRequest() {
			var bytes = new byte[] { 1, 2, 3 };
			var length = 3;

			var result = service.buildChunkRequest().apply(bytes, length);

			assertThat(result.getAttachments().getContent().toByteArray()).isEqualTo(bytes);
		}
	}

	@Nested
	class TestBuildAttachmentMetadataMapper {

		private static final BayernIdAttachment ATTACHMENT = BayernIdAttachmentTestFactory.create();

		@Mock
		private GrpcSendBayernIdMessageRequest attachmentMetadataRequest;

		@Test
		void shouldCallMapper() {
			service.buildAttachmentMetadataMapper().apply(ATTACHMENT);

			verify(mapper).toSendBayernIdAttachmentsMetadataRequest(ATTACHMENT);
		}

		@Test
		void shouldReturnResult() {
			when(mapper.toSendBayernIdAttachmentsMetadataRequest(any())).thenReturn(attachmentMetadataRequest);

			var result = service.buildAttachmentMetadataMapper().apply(ATTACHMENT);

			assertThat(result).isEqualTo(attachmentMetadataRequest);
		}
	}

	@Nested
	class TestWaitUntilTransferCompleted {

		@Mock
		private MessageWithFilesSender sender;
		@Mock
		private GrpcSendBayernIdMessageResponse grpcResponse;
		@Mock
		private CompletableFuture<GrpcSendBayernIdMessageResponse> resultFuture;

		@BeforeEach
		void setup() {
			doReturn(resultFuture).when(sender).getResultFuture();
		}

		@Test
		void shouldCallGet() {
			service.waitUntilTransferCompleted(sender);

			verify(sender).getResultFuture();
		}

		@Test
		@SneakyThrows
		void shouldReturnResult() {
			when(resultFuture.get(anyLong(), any())).thenReturn(grpcResponse);

			var result = service.waitUntilTransferCompleted(sender);

			assertThat(result).isEqualTo(grpcResponse);
		}

		@SneakyThrows
		@Test
		void shouldHandleInterruptedException() {
			var interruptedException = new InterruptedException();

			when(resultFuture.get(anyLong(), any())).thenThrow(interruptedException);

			assertThrows(TechnicalException.class, () -> service.waitUntilTransferCompleted(sender));
			verify(sender).cancelOnError(interruptedException);
		}

		@Test
		@SneakyThrows
		void shouldHandleTimeout() {
			when(resultFuture.get(anyLong(), any())).thenThrow(new TimeoutException());

			assertThrows(TechnicalException.class, () -> service.waitUntilTransferCompleted(sender));
			verify(sender).cancelOnTimeout();
		}

		@Test
		@SneakyThrows
		void shouldHandleExecutionException() {
			var executionException = new ExecutionException(new RuntimeException());
			when(resultFuture.get(anyLong(), any())).thenThrow(executionException);

			assertThrows(TechnicalException.class, () -> service.waitUntilTransferCompleted(sender));
			verify(sender).cancelOnTimeout();
		}

	}

	@Nested
	class TestGetPostfach {

		@Test
		void shouldHasTyp() {
			var postfachType = service.getPostfachType();

			assertThat(postfachType).isEqualTo(BayernIdPostfachRemoteService.POSTFACH_TYPE);
		}

		@Test
		void shouldSetReplyNotAllowed() {
			var isReplyAllowed = service.isReplyAllowed();

			assertThat(isReplyAllowed).isFalse();
		}
	}
}